import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../'))
import pickle
import gdown
import numpy as np
import pkg_resources
import tensorflow_datasets as tfds
from utils import utils
from nkia.ml.cnn_nlp_model import CNN


class ClassifyProducts(object):

    def load_model_to_memory(self, model_name):
        model_version = pkg_resources.get_distribution('nkia').version
        os.makedirs(model_name) if model_name not in os.listdir() else ...
        utils.get_model_files_by_version(model_name + '/', model_version)

        tokenizer = tfds.features.text.SubwordTextEncoder.load_from_file(
            './'+model_name+'/'+model_version+'/token')

        nb_classes = 29 if model_name == 'category_model' else 2

        cnn = CNN(vocab_size=tokenizer.vocab_size, emb_dim=300, nb_filters=100,
                  ffn_units=256, nb_classes=nb_classes, dropout_rate=0.2)

        cnn.load_weights('./'+model_name+'/'+model_version +
                         '/model_checkpoint/cp.ckpt').expect_partial()
        return cnn

    def load_dictionary(self):
        if 'nlp_dictionary' not in os.listdir():
            os.makedirs('nlp_dictionary')
            gdown.download('https://project-elements-nk.s3.amazonaws.com/dictionary.pkl',
                           'nlp_dictionary/dictionary.pkl', quiet=False)

        return utils.instantiate_spelling_corrector('nlp_dictionary/dictionary.pkl')

    def inference_from_category_model(self, model, sym_spell, product_name=[''], description=['']):
        """
        This method is responsible to infer the product category based on your caracteristics. Pass just one product
        by time.

        Args:
            product_name (list):  ['energético sabor baunilha']
            description (list):  ['Esse produto é simplesmente fantástico']

        Returns:
            str: Category name. e.g: 'Bebidas não alcoolicas'

        """
        X = self.preprocessing_fn(
            sym_spell, product_name, description, 'category_model')

        product_category = model(np.array(X), training=False).numpy()[0]
        product_category_proba = product_category[product_category.argmax()]

        if product_category_proba < 0.6:
            return 'categoria desconhecida'

        return self.label_encoder.inverse_transform([product_category.argmax()])[0]

    def inference_from_food_model(self, model, sym_spell, product_name=[''], description=['']):
        """
        This method is responsible to infer if a product is a food or not. Pass just one product
        by time.

        Args:
            product_name (list):  ['energético sabor baunilha']

        Returns:
            str: food or not food.

        """
        X = self.preprocessing_fn(
            sym_spell, product_name, description, 'food_model')

        product_type = model(np.array(X), training=False).numpy()

        if product_type > 0.4 and product_type < 0.6:
            return 'sem certeza'

        product_type = 1 if product_type >= 0.6 else 0

        return self.label_encoder.inverse_transform([product_type])[0]

    def preprocessing_fn(self, sym_spell, product_name, description, model_name):
        model_version = pkg_resources.get_distribution('nkia').version
        tokenizer = tfds.features.text.SubwordTextEncoder.load_from_file(
            './'+model_name+'/'+model_version+'/token')

        X = utils.get_cleaned_predictor(
            sym_spell, [description], [product_name], model_name)

        tokenized_predictor = [tokenizer.encode(sentence) for sentence in X]

        self.label_encoder = self.get_label_encoder(
            './'+model_name+'/'+model_version+'/label_encoder.pkl')

        return tokenized_predictor

    def get_label_encoder(self, encoder_path):
        pkl_encoder_file = open(encoder_path, 'rb')
        encoder = pickle.load(pkl_encoder_file)
        pkl_encoder_file.close()
        return encoder


ClassifyProducts()



import boto3


class DeployModel(object):

    def deploy_to_s3(self, model_version):
        s3_resource = boto3.resource('s3')
        bucket_name = 'category-model' if 'category_model' in model_version.split('/') else 'food-model'

        print('Uploading {} to Amazon S3 bucket {}'.format(model_version, bucket_name))
        s3_resource.Object(bucket_name, model_version.split('/')[-1]).put(
            Body=open(model_version, 'rb'), ACL='public-read')



from pathlib import Path
import os
import sys
import re
import string
import tarfile
import unidecode
import nltk
import tensorflow_datasets as tfds
import tensorflow as tf
import gdown
from symspellpy import SymSpell, Verbosity
import numpy as np


def get_cleaned_predictor(sym_spell, description, product_name, model_name):
    stop_words = get_stop_words_list(model_name)
    description = [get_description_and_name(x, stop_words) for x in description]
    product_name = [get_description_and_name(x, stop_words) for x in product_name]

    X = join_all_predictors(description, product_name)

    X = correct_word_spelling(sym_spell, X)
    return X


def get_stop_words_list(model_name):
    stop_words = nltk.corpus.stopwords.words('portuguese')

    custom_stop_words = ['carrefour', 'compre', 'aproveite', 'produto', 'qualidade', 'oportunidade', 'preco', 'justo', 
    'voce', 'curtiu', 'qualidade', 'alem', 'disso']

    if model_name == 'category_model':
        food_model_stop_words = ['unidades', 'saboroso', 'boca', 'g', 'ml', 'cm', 'kg', 'sabor', 'kcal', 'mg', 'delicioso', 'mercado']
        custom_stop_words.extend(food_model_stop_words)
    
    stop_words.extend(custom_stop_words)
    return stop_words

def get_description_and_name(data, stop_words):
    data = unidecode.unidecode(data[0]).lower()
    data = re.sub(r'['+string.punctuation+']', '', data)
    data = [word for word in data.split() if word not in stop_words]

    return ' '.join(data)


def join_all_predictors(description, product_name):
    count = 0
    X = []

    for name in product_name:
        X.append(name + ' ' + description[count])
        count += 1

    return X


def preprocessing_fn(X, token_path):
    X, tokenizer = tokenize_data(X, token_path)
    X = padding_matrix(X)

    print('Preprocessing_fn finished')
    sys.stdout.flush()
    return X, tokenizer


def tokenize_data(data, token_path):
    tokenizer = tfds.features.text.SubwordTextEncoder.build_from_corpus(
        data, target_vocab_size=2**14)

    tokenizer.save_to_file(token_path)
    data = [tokenizer.encode(sentence) for sentence in data]

    return data, tokenizer


def padding_matrix(data):
    max_sentence_len = max([len(sentence) for sentence in data])
    data = tf.keras.preprocessing.sequence.pad_sequences(
        data,
        value=0,
        padding='post',
        maxlen=max_sentence_len)

    return data


def read_file(file_name, content=''):
    if content:
        with open(os.path.join(Path(os.path.dirname(__file__)), file_name), 'w') as _file:
            _file.write(str(content))
            _file.close()
    else:
        with open(os.path.join(Path(os.path.dirname(__file__)), file_name)) as _file:
            return _file.read()


def get_model_files_by_version(model_name, model_version):
    splitted_name = model_name.split('/')
    bucket = 'category-model' if 'category_model' in splitted_name else 'food-model'

    url = 'https://{}.s3.amazonaws.com/{}'.format(
        bucket, model_version + '.tar.gz')

    if model_version not in os.listdir(model_name):
        download_model(url, model_name + model_version + '.tar.gz')


def download_model(url, output):
    gdown.download(url, output, quiet=False)

    tar = tarfile.open(output)
    tar.extractall('/'.join(output.split('/')[0:-1]))
    tar.close()
    os.remove(output)


def instantiate_spelling_corrector(dict_path='./src/utils/datasets/dictionary.pkl'):
    sym_spell = SymSpell(max_dictionary_edit_distance=2, prefix_length=7)
    sym_spell.load_pickle(dict_path)
    return sym_spell


def correct_word_spelling(sym_spell, corpus):
    final_corpus = []
    corpus = np.char.split(np.array(corpus))

    for phrase in corpus:
        new_phrase = [get_word_suggestion(sym_spell, x) for x in phrase]
        final_corpus.append(' '.join(new_phrase).replace('  ', ' ').strip())

    return final_corpus


def get_word_suggestion(sym_spell, input_term):
    suggestion = sym_spell.lookup(
        input_term, Verbosity.TOP, max_edit_distance=2)

    if len(suggestion) > 0:
        return suggestion[0].term

    return ''
